
/*
 * flow: a Smalltalk external streaming framework
 * version 2a2
 *
 * flow.c - module initialization and utilities
 * version 2
 *
 * Craig Latta
 * netjam.org/flow
 */

#include "flow.h"

/*
 * utilities
 */

void startIP(void) {
  /* No action is required here for platforms other than win32. */
#ifdef WIN32
  WSADATA	wsadata;
  WORD	wVersionRequired = 0x0002;

  if (WSAStartup(wVersionRequired, &wsadata) != 0) vm->primitiveFail();
#endif
}

#ifndef _WIN32_WCE
void announceAvailableMIDIInput(void *ignored) {
  int		index;
  midiPort	*port;

#ifdef WIN32
  while (1) {
    WaitForSingleObject(midiInputAvailable, INFINITE);
    for (index = 0; index < MAXMIDIPORTS; index++) {
      port = flowMidiPorts[index];
      if (port && (port->incomingShortMessagesStartIndex != port->incomingShortMessagesStopIndex))
	vm->signalSemaphoreWithIndex(port->shortMessageAvailability);}}
#endif
}
#endif

#ifndef _WIN32_WCE
void announceAvailableSysEx(void *ignored) {
  int		index;
  midiPort	*port;

#ifdef WIN32
  while (1) {
    WaitForSingleObject(sysExAvailable, INFINITE);
    for (index = 0; index < MAXMIDIPORTS; index++) {
      port = flowMidiPorts[index];
      if (port &&
	  (port->incomingSystemExclusiveMessagesStartIndex != port->incomingSystemExclusiveMessagesStopIndex))
	vm->signalSemaphoreWithIndex(port->systemExclusiveMessageAvailability);}}
#endif
}
#endif

#ifndef _WIN32_WCE
void startMIDI(void) {
  /* Only win32 is supported currently. */
#ifdef WIN32
  DWORD threadID;

  midiInputAvailable = (CreateEvent(NULL, 0, 0, NULL));

  availableMIDIInputAnnouncement =
    (CreateThread(NULL, 0,
		  (LPTHREAD_START_ROUTINE) &announceAvailableMIDIInput,
		  NULL, 0, &threadID));

  sysExAvailable =
    (CreateEvent(NULL, 0, 0, NULL));

  availableSysExAnnouncement =
    (CreateThread(NULL, 0,
		  (LPTHREAD_START_ROUTINE) &announceAvailableSysEx,
		  NULL, 0, &threadID));
#endif
}
#endif

void stopMIDI(void) {
#ifndef _WIN32_WCE
  int portIndex;

  /* Close all open ports. */
  for (portIndex = 0; portIndex < MAXMIDIPORTS; portIndex++) closeMIDIPort(portIndex);
  TerminateThread(availableMIDIInputAnnouncement, 0);
  CloseHandle(midiInputAvailable);
  TerminateThread(availableSysExAnnouncement, 0);
  CloseHandle(sysExAvailable);
#endif
}

int addressForStackValue(int index) {
  return *((int *) (vm->stackObjectValue(index) + ObjectHeaderSize));}

void writeNewResourceHandle(int size) {
  int	handleOop = vm->stackObjectValue(0);
  int	address;

  address = (int) calloc(1, size);
  memcpy((void *) (handleOop + ObjectHeaderSize),
	 (const void *) &address,
	 4);
  vm->pop(1);}

int startThread(threadSync *sync, void *function, void *parameter) {
#ifdef UNIX
  pthread_t	thread;
  pthread_cond_t	request = PTHREAD_COND_INITIALIZER;
  pthread_mutex_t	mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
#ifdef WIN32
  DWORD		dwThreadID;
  HANDLE		hThread;
#endif

#ifdef UNIX
  sync->requestAlreadyOccurred = False;
  sync->thread = thread;
  sync->request = request;
  sync->mutex = mutex;

  /*
   * Apparently, in POSIX, one may meaningfully set the priority
   * only of threads with superuser privileges.
   */
  pthread_create(&sync->thread,
		 NULL,
		 function,
		 parameter);
#endif
#ifdef WIN32
  sync->pendingEvent = 
    CreateEvent(NULL,	/* no security */
		False,	/* can implicitly reset */
		False,	/* initial pendingEvent reset */
		NULL);	/* no name */

  if (NULL == (hThread =
	       CreateThread((LPSECURITY_ATTRIBUTES) NULL,
			    0,
			    (LPTHREAD_START_ROUTINE) function,
			    (LPVOID) parameter,
			    0,
			    &dwThreadID))) {
    return False;}

  SetThreadPriority(hThread, THREAD_PRIORITY_HIGHEST);
  sync->thread = hThread;
#endif
  return True;}

int startScribingThreads(netResource	*resource,
			 void		*writingFunction,
			 void		*readingFunction,
			 void		*parameter) {
  if (!startThread(&resource->writing.sync,
		   writingFunction,
		   parameter))
    return False;
  else {
    if (!startThread(&resource->reading.sync,
		     readingFunction,
		     parameter))
      return False;
    else return True;}}

void waitForThreadSignal(threadSync *sync) {
#ifdef UNIX
	pthread_mutex_lock(&sync->mutex);
	while(!sync->requestAlreadyOccurred) {
		pthread_cond_wait(&sync->request, &sync->mutex);}
	sync->requestAlreadyOccurred = False;
	pthread_mutex_unlock(&sync->mutex);
#endif
#ifdef WIN32
	WaitForSingleObject(sync->pendingEvent, INFINITE);
#endif
}

void signalThread(threadSync *sync) {
#ifdef UNIX
  pthread_mutex_lock(&sync->mutex);
  sync->requestAlreadyOccurred = True;
  pthread_cond_signal(&sync->request);
  pthread_mutex_unlock(&sync->mutex);
#endif
#ifdef WIN32
  SetEvent(sync->pendingEvent);
#endif
}

void signalSynchronizedResourceThread(thread *thread) {
  thread->timeout = vm->stackValue(0);
  if (thread->timeout == CurrentMemory->NilObject())
    thread->timeout = -1;
  else if (thread->timeout < 0) thread->timeout = 0;
  thread->operation = vm->stackIntegerValue(1);
  signalThread(&thread->sync);
  vm->pop(3);}

void stopThread(threadSync *sync) {
  signalThread(sync);

  /* Wait until the thread has been halted. */
#ifdef UNIX
  pthread_join(sync->thread, NULL);
#endif
#ifdef WIN32
  WaitForSingleObject(sync->thread, INFINITE);
  CloseHandle(sync->thread);
  CloseHandle(sync->pendingEvent);
  sync->pendingEvent = NULL;
#endif
}

void killThread(threadSync *sync) {
#ifdef UNIX
  pthread_cancel(sync->thread);
#endif
#ifdef WIN32
  TerminateThread(sync->thread, 0);
  CloseHandle(sync->thread);
  CloseHandle(sync->pendingEvent);
  sync->pendingEvent = NULL;
#endif
}

void stopIP(void) {
#ifdef WIN32
  WSACleanup();
#endif
}

/* primitives */

void associateWithReadabilityIndexAndWritabilityIndex(void) {
  /*
   * associate: netResourceHandle
   * withReadabilityIndex: readabilityIndex
   * andWritabilityIndex: writabilityIndex
   */

  netResource	*netResourcePointer = (netResource *)(addressForStackValue(2));

  if (vm->stackObjectValue(2) == vm->nilObject()) {
    vm->primitiveFail();
    return False;}

  netResourcePointer->reading.sync.semaphore = vm->stackIntegerValue(1);
  netResourcePointer->writing.sync.semaphore = vm->stackIntegerValue(0);

  ;}
